# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from datetime import timedelta
from enum import Enum
from pydantic import BaseModel, computed_field, ConfigDict, Field, field_serializer, field_validator

# TODO: Wire up the data_source field to poll scheduling (everything is currently short-poll because this isn't used).
# TODO: Should NEVER actually be an option? Could it just be None?
class DataSource(Enum):
    SHORT_POLL = "short_poll"
    LONG_POLL = "long_poll"
    NEVER = "never"
    POLL_ONCE = "poll_once"
    STATIC = "static"


class EquipmentConfig(BaseModel):
    model_config = ConfigDict(validate_assignment=True, populate_by_name=True)
    active: bool | None = None
    group: str | None = None
    # TODO: If this needs to be an int, we may need to use milliseconds someplace.
    polling_interval: int | None = Field(default=None, alias='interval')
    publish_single_depth: bool | None = Field(default=None, alias='publish_depth_first_single')
    publish_single_breadth: bool | None = Field(default=None, alias='publish_breadth_first_single')
    publish_multi_depth: bool | None = Field(default=None, alias='publish_depth_first_multi')
    publish_multi_breadth: bool | None = Field(default=None, alias='publish_breadth_first_multi')
    publish_all_depth: bool | None = Field(default=None, alias='publish_depth_first_all')
    publish_all_breadth: bool | None = Field(default=None, alias='publish_breadth_first_all')
    reservation_required_for_write: bool = False

    @field_validator('polling_interval', mode='before')
    @classmethod
    def _normalize_polling_interval(cls, v):
        # TODO: This does not match int above, but we may need to convert to ms in calculations.
        return None if v == '' or v is None else float(v)

class PointConfig(EquipmentConfig):
    data_source: DataSource = Field(default=DataSource.SHORT_POLL, alias='Data Source')
    notes: str = Field(default='', alias='Notes')
    reference_point_name: str = Field(default='', alias='Reference Point Name')
    stale_timeout_configured: float | None = Field(default=None, alias='stale_timeout')
    stale_timeout_multiplier: float = Field(default=3)
    units: str = Field(default='', alias='Units')
    units_details: str = Field(default='', alias='Unit Details')
    volttron_point_name: str = Field(alias='Volttron Point Name')
    writable: bool = Field(default=False, alias='Writable')

    @field_validator('data_source', mode='before')
    @classmethod
    def _normalize_data_source(cls, v):
        # TODO: This never converts to DataSource.
        # TODO: Data Source enum needs something to tell Data Point how to serialize it, otherwise enable/disable will fail.
        return v.lower()

    @field_serializer('data_source')
    def _serialize_data_source(self, data_source):
        return data_source.value

    @computed_field
    @property
    def stale_timeout(self) -> timedelta | None:
        if self.stale_timeout_configured is None and self.polling_interval is None:
            return None
        else:
            return timedelta(seconds=(self.stale_timeout_configured
                    if self.stale_timeout_configured is not None
                    else self.polling_interval * self.stale_timeout_multiplier))

    @stale_timeout.setter
    def stale_timeout(self, value):
        self.stale_timeout_configured = value


class DeviceConfig(EquipmentConfig):
    all_publish_interval: float = 0.0
    allow_duplicate_remotes: bool = False
    equipment_specific_fields: dict = {}
    registry_config: list[PointConfig] = []


class RemoteConfig(BaseModel):
    model_config = ConfigDict(extra='allow', validate_assignment=True)
    debug: bool = False
    driver_type: str
    heart_beat_point: str | None = None  # TODO: This needs to become a set (multiple devices could have multiple points).
    module: str | None = None


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
import logging
import resource

from contextlib import contextmanager
from gevent.lock import BoundedSemaphore, DummySemaphore


_log = logging.getLogger(__name__)

_socket_lock = None

def get_system_socket_limit():
    # Increase open files resource limit to max or 8192 if unlimited
    system_socket_limit = None
    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    except OSError:
        _log.exception('error getting open file limits')
    else:
        if soft != hard and soft != resource.RLIM_INFINITY:
            try:
                system_socket_limit = 8192 if hard == resource.RLIM_INFINITY else hard
                resource.setrlimit(resource.RLIMIT_NOFILE, (system_socket_limit, hard))
            except OSError:
                _log.exception('error setting open file limits')
            else:
                _log.debug('open file resource limit increased from %d to %d', soft,
                           system_socket_limit)
        if soft == hard:
            system_socket_limit = soft
    return system_socket_limit

def setup_socket_lock(max_open_sockets=None):
    if max_open_sockets is not None:
        max_connections = int(max_open_sockets)
        _log.info("maximum concurrently open sockets limited to " + str(max_open_sockets))
    else:
        system_socket_limit = get_system_socket_limit()
        if system_socket_limit is not None:
            max_connections = int(system_socket_limit * 0.8)
            _log.info("maximum concurrently open sockets limited to " + str(max_open_sockets)
                      + " (derived from system limits)")
        else:
            max_connections = 0
            _log.warning(
                "No limit set on the maximum number of concurrently open sockets. "
                "Consider setting max_open_sockets if you plan to work with 800+ modbus devices."
            )
    configure_socket_lock(max_connections)

def configure_socket_lock(max_connections=0):
    global _socket_lock
    if _socket_lock is not None:
        raise RuntimeError("socket_lock already configured!")
    if max_connections < 1:
        _socket_lock = DummySemaphore()
    else:
        _socket_lock = BoundedSemaphore(max_connections)


@contextmanager
def socket_lock():
    global _socket_lock
    if _socket_lock is None:
        raise RuntimeError("socket_lock not configured!")
    _socket_lock.acquire()
    try:
        yield
    finally:
        _socket_lock.release()


_publish_lock = None


def configure_publish_lock(max_connections=0):
    if max_connections < 1:
        _log.warning(
            "No limit set on the maximum number of concurrent driver publishes. "
            "Consider setting max_concurrent_publishes if you plan to work with many devices."
        )
    else:
        _log.info("maximum concurrent driver publishes limited to " + str(max_connections))
    global _publish_lock
    if _publish_lock is not None:
        raise RuntimeError("socket_lock already configured!")
    if max_connections < 1:
        _publish_lock = DummySemaphore()
    else:
        _publish_lock = BoundedSemaphore(max_connections)


@contextmanager
def publish_lock():
    global _publish_lock
    if _publish_lock is None:
        raise RuntimeError("socket_lock not configured!")
    _publish_lock.acquire()
    try:
        yield
    finally:
        _publish_lock.release()



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import gevent
import logging
import random

from volttron.client.messaging import headers as headers_mod
from volttron.client.vip.agent.errors import Again, VIPError
from volttron.utils import format_timestamp, get_aware_utc_now

from volttron.driver.base.driver_locks import publish_lock


_log = logging.getLogger(__name__)


def publication_headers():
    # TODO: Sync_Timestamp won't work, so far, because time_slot_offset assumed the device was polled once per round.
    #  Since we are polling through a hyperperiod that may include multiple rounds for a given point or equipment,
    #  this no longer makes sense. Also, what if some points are polled multiple times compared to others?
    #  CAN SCHEDULED ALL_PUBLISHES REPLACE THIS MECHANISM IF THEY ARE GENERATED ALL AT ONCE?
    #  OR JUST USE THIS IS ALL-TYPE PUBLISHES ON A SCHEDULE?
    utcnow_string = format_timestamp(get_aware_utc_now())
    headers = {
        headers_mod.DATE: utcnow_string,
        headers_mod.TIMESTAMP: utcnow_string,
        # headers_mod.SYNC_TIMESTAMP: format_timestamp(current_start - timedelta(seconds=self.time_slot_offset))
    }
    return headers

def publish_wrapper(vip, topic, headers, message):
    while True:
        try:
            with publish_lock():
                _log.debug("publishing: " + topic)
                # TODO: Do we really need to block on every publish call?
                vip.pubsub.publish('pubsub', topic, headers=headers, message=message).get(timeout=10.0)
                _log.debug("finish publishing: " + topic)
        except gevent.Timeout:
            _log.warning("Did not receive confirmation of publish to " + topic)
            break
        except Again:
            _log.warning("publish delayed: " + topic + " pubsub is busy")
            gevent.sleep(random.random())
        except VIPError as ex:
            _log.warning("driver failed to publish " + topic + ": " + str(ex))
            break
        else:
            break


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import gevent
import logging
import traceback

from collections import defaultdict
from typing import Any, cast
from weakref import WeakSet


from volttron.client.vip.agent import Agent

from volttron.driver.base.interfaces import BaseInterface
from volttron.driver.base.config import PointConfig, RemoteConfig
from volttron.driver.base.utils import publication_headers, publish_wrapper
# from platform_driver.poll_scheduler import PollSet  # TODO: This should not import from driver. Need to relocate PollSet.

_log = logging.getLogger(__name__)


class DriverAgent:
    def __init__(self, config: RemoteConfig, core, equipment_model, scalability_test, tz: str, unique_id: Any,
                 vip: Agent.Subsystems):
        self.config: RemoteConfig = config
        self.core = core
        self.equipment_model = equipment_model  # TODO: This should probably move out of the agent and into the base or a library.
        self.scalability_test = scalability_test  # TODO: If this is used from here, it should probably be in the base driver.
        self.tz: str = tz  # TODO: This won't update here if it is updated in the agent. Can it's use be moved out of here?
        self.unique_id: Any = unique_id
        self.vip: Agent.Subsystems = vip
        
        # State Variables
        self.equipment = WeakSet()
        self.heart_beat_value = 0
        self.interface = None
        self.publishers = {}

        try:
            # TODO: What happens if we have multiple device nodes on this remote?
            #  Are we losing all settings but the first?
            klass = BaseInterface.get_interface_subclass(self.config.driver_type)
            interface = klass(self.config, self)
            self.interface = cast(BaseInterface, interface)
        except ValueError as e:
            _log.error(f"Failed to setup device: {e}")
            raise e

    def add_registers(self, registry_config: list[PointConfig], base_topic: str):
        """
        Configure a set of registers on this remote.

        :param registry_config: A list of registry points represented as PointConfigs
        :param base_topic: The portion of the topic shared by all points in this registry.
        """
        _log.debug(f'IN ADD REGISTERS WITH REGISTRY_CONFIG: {registry_config}')
        for register_config in registry_config:
            register = self.interface.create_register(register_config)
            self.interface.insert_register(register, base_topic)
        # TODO: Finalize method is only used by bacnet, currently, and that pauses 30s on each device if it can't connect.
        try:
            self.interface.finalize_setup(initial_setup=True)
        except BaseException as e:
            _log.warning(f'Exception occurred while finalizing setup of interface for {self.unique_id}: {e}.')

        for point_name in self.interface.get_register_names():
            register = self.interface.get_register_by_name(point_name)
            point = self.equipment_model.get_node(point_name)
            # TODO: It might be more reasonable to either have the register be aware of the type mappings or have a
            #  type-mapping function separately. This is rather limiting. What is "ts" anyway? TypeScript?
            if register.register_type == 'bit':
                ts_type = 'boolean'
            else:
                if register.python_type is int:
                    ts_type = 'integer'
                elif register.python_type is float:
                    ts_type = 'float'
                elif register.python_type is str:
                    ts_type = 'string'
            # TODO: Why is there not an else here? ts_type may be undefined.
            # TODO: meta_data may belong in the PointNode object. This function could take points instead of their
            #  configs and pack the data into the PointNode instead of a separate dictionary in this class.
            point.meta_data = {
                'units': register.get_units(),
                'type': ts_type,
                'tz': self.tz
            }

    def poll_data(self, poll_set): # PollSet):
        _log.debug(f'@@@@@ Polling: {self.unique_id}')
        if self.scalability_test:  # TODO: Update scalability testing.
            self.scalability_test.poll_starting(self.unique_id)
        try:
            _log.debug('@@@@@ BEFORE GET_MULTIPLE_POINTS IN POLL_DATA')
            results, errors = self.interface.get_multiple_points(poll_set.points.keys())
            _log.debug('@@@@@ AFTER GET_MULTIPLE_POINTS IN POLL_DATA')
            for failed_point, failure_message in errors.items():
                _log.warning(f'Failed to poll {failed_point}: {failure_message}')
            if results:
                for topic, value in results.items():
                    point = poll_set.points.get(topic)
                    if point and point.active:
                        point.last_value = value
                self.publish_poll(results, poll_set)
            return True  # TODO: There could really be better logic in the method to measure success.
        except (Exception, gevent.Timeout) as e:
            _log.error(f'Exception while polling {self.unique_id}: {e}')
            if self.config.debug:
                # TODO: Add an RPC to turn on debugging for individual remotes. Maybe for nodes as well?
                tb = traceback.format_exc()
                _log.error(tb)
            return False
        finally:
            if self.scalability_test:
                self.scalability_test.poll_ending(self.unique_id)

    # noinspection DuplicatedCode
    def publish_poll(self, results, poll_set):
        headers = publication_headers()
        for point_topic in poll_set.single_depth:
            if point_topic in results:
                try:
                    publish_wrapper(self.vip, point_topic, headers=headers, message=[
                        results[point_topic], self.equipment_model.get_node(point_topic).meta_data
                    ])
                except Exception as e:
                    _log.warning(f'Failed to publish single_depth point_topic: {point_topic} -- {e}')
        for point_topic, publish_topic in poll_set.single_breadth:
            if point_topic in results:
                try:
                    publish_wrapper(self.vip, publish_topic, headers=headers, message=[
                        results[point_topic], self.equipment_model.get_node(point_topic).meta_data
                    ])
                except Exception as e:
                    _log.warning(f'Failed to publish single_breadth publish_topic: {publish_topic}'
                                 f' for identifier: {point_topic} -- {e}')
        for device_topic, points in poll_set.multi_depth.items():
            try:
                publish_wrapper(self.vip, f'{device_topic}/multi', headers=headers, message=[
                    {point.rsplit('/', 1)[-1]: results[point] for point in points if point in results},
                    {point.rsplit('/', 1)[-1]: self.equipment_model.get_node(point).meta_data for point in points}
                ])
            except Exception as e:
                _log.warning(f'Failed to publish multi_depth device_topic: {device_topic} -- {e}')
        for publish_topic, points in poll_set.multi_breadth.items():
            try:
                publish_wrapper(self.vip, f'{publish_topic}/multi', headers=headers, message=[
                    {point.rsplit('/', 1)[-1]: results[point] for point in points if point in results},
                    {point.rsplit('/', 1)[-1]: self.equipment_model.get_node(point).meta_data for point in points}
                ])
            except Exception as e:
                _log.warning(f'Failed to publish multi_breadth publish_topic: {publish_topic} -- {e}')

    def heart_beat(self):
        if self.config.heart_beat_point is None:
            return
        self.heart_beat_value = int(not bool(self.heart_beat_value))
        # TODO: config.heart_beat_point should be a set.
        self.set_point(self.config.heart_beat_point, self.heart_beat_value)

    def get_point(self, topic, **kwargs):
        return self.interface.get_point(topic, **kwargs)

    def set_point(self, topic, value, **kwargs):
        return self.interface.set_point(topic, value, **kwargs)

    def get_multiple_points(self, topics, **kwargs):
        return self.interface.get_multiple_points(topics, **kwargs)

    def set_multiple_points(self, topics_values, **kwargs):
        return self.interface.set_multiple_points(topics_values, **kwargs)

    def revert_point(self, topic, **kwargs):
        self.interface.revert_point(topic, **kwargs)

    def revert_all(self, **kwargs):
        self.interface.revert_all(**kwargs)

    def publish_push(self, results):
        et = self.equipment_model
        headers = publication_headers()
        multi_depth_values, multi_depth_meta, multi_breadth_values, multi_breadth_meta = (
            defaultdict(dict), defaultdict(dict), defaultdict(dict), defaultdict(dict))
        for point_topic, value in results.items():
            point_depth_topic, point_breadth_topic = et.get_point_topics(point_topic)
            device_depth_topic, device_breadth_topic = et.get_device_topics(point_topic)
            point_node = self.equipment_model.get_node(point_topic)
            if et.is_published_single_depth(point_topic):
                publish_wrapper(self.vip, point_depth_topic, headers, [value, point_node.meta_data])
            if et.is_published_single_breadth(point_topic):
                publish_wrapper(self.vip, point_breadth_topic, headers, [value, point_node.meta_data])
            if et.is_published_multi_depth(point_topic):
                point_name = point_topic.rsplit('/', 1)[-1]
                multi_depth_values[device_depth_topic][point_name] = value
                multi_depth_meta[device_depth_topic][point_name] = point_node.meta_data
            if et.is_published_multi_breadth(point_topic):
                point_name = point_topic.rsplit('/', 1)[-1]
                multi_breadth_values[device_breadth_topic][point_name] = value
                multi_breadth_meta[device_breadth_topic][point_name] = point_node.meta_data
        if multi_depth_values:
            for device_topic in multi_depth_values:
                publish_wrapper(self.vip, device_topic, headers,
                                      [multi_depth_values[device_topic], multi_depth_meta[device_topic]])
        if multi_breadth_values:
            for device_topic in multi_breadth_values:
                publish_wrapper(self.vip, device_topic, headers,
                                [multi_breadth_values[device_topic], multi_breadth_meta[device_topic]])

    def add_equipment(self, device_node):
        # TODO: Is logic needed for scheduling or any other purpose on adding equipment to this remote?
        _log.debug(f'IN ADD EQUIPMENT, with device_node: {device_node.identifier}')
        self.add_registers([p.config for p in self.equipment_model.points(device_node.identifier)],
                           device_node.identifier)
        self.equipment.add(device_node)

    @property
    def point_set(self):
            return {point for equip in self.equipment for point in self.equipment_model.points(equip.identifier)}
